#pragma once

struct MeshVertex
{
	static ulong decl[];

	oVector3 p, n;
	oVector2 t;
};

oVector3 GetBoxGeomLength(const dGeomID& geom) throw();

class Mesh
{
protected:
	struct IndexedPrimitive 
	{
		D3DPRIMITIVETYPE type;
		UINT minIndex;
		UINT numVertices;
		UINT startIndex;
		UINT primitiveCount;
	};

	ushort numVertices, numIndices, numPrimitives;

	oHeapPtr<MeshVertex>::Type vertices;
	oHeapPtr<ushort>::Type indices;
	oHeapPtr<IndexedPrimitive>::Type primitives;

	inline void Allocate(const ushort& numVertices_, const ushort& numIndices_, const ushort& numPrimitives_);
	inline void CreateBoxFace(MeshVertex* const& buffer, const oVector3& normal, const oReal& u, const oReal& v) throw();

public:
	Mesh();

	void CreateBox(const oVector3& length);
	void CreateBox(const oVector3& length, const oVector3 textureScale);
	void CreateCylinder(const oReal& radius, const oReal& length, const ushort& segments);
	void CreateGroundPlane(const oVector2& extents);
	void CreateSphere(const oReal& radius, const ushort& segments, const ushort& levels);

	const ushort& GetVertexCount() const throw() { return numVertices; }
	const ushort& GetIndexCount() const throw() { return numIndices; }

	void GetVertices(MeshVertex* const& buffer, const oMatrix4x4& meshTransform, const oMatrix4x4& normalTransform) const throw();
	void GetIndices(ushort* const& buffer, const ushort& offset) const throw();

	void DrawIndexedPrimitive(IDirect3DDevice8* const& device, const ulong& vertexOffset, const ulong& indexOffset) const throw();
};

class GeomMesh
{
protected:
	dGeomID geom;
	const Mesh* mesh;

	ushort vertexOffset, indexOffset;

public:
	GeomMesh();
	
	GeomMesh(const dGeomID& geom_, const Mesh* const& mesh_);
	void Create(const dGeomID& geom_, const Mesh* const& mesh_);

	template<class Traits> 
	ushort GetVertices(oVertexBuffer<MeshVertex, Traits>& vertices, const ushort& vertexOffset_)
	{
		vertexOffset = vertexOffset_;
		UpdateVertices(vertices);
		return vertexOffset + mesh->GetVertexCount();
	}

	template<class Traits> 
	ushort GetIndices(oIndexBuffer<ushort, Traits>& indices, const ushort& indexOffset_)
	{
		indexOffset = indexOffset_;
		ushort* const buffer = indices.Lock(indexOffset, mesh->GetIndexCount());
		mesh->GetIndices(buffer, vertexOffset);
		indices.Unlock();
		return indexOffset + mesh->GetIndexCount();
	}

	template<class Traits> 
	void UpdateVertices(oVertexBuffer<MeshVertex, Traits>& vertices)
	{
		// compute the mesh and normal transform
		oMatrix4x4 normalTransform(dGeomGetRotation(geom));
		oMatrix4x4 meshTransform(normalTransform);
		meshTransform.PostTranslate(oVector3(dGeomGetPosition(geom)));
	   
		// get the vertices
		MeshVertex* const buffer = vertices.Lock(vertexOffset, mesh->GetVertexCount());
		mesh->GetVertices(buffer, meshTransform, normalTransform);
		vertices.Unlock();
	}

	void DrawIndexedPrimitive(IDirect3DDevice8* const& device);
};