#pragma once

void ComputeShadowVolumeBasis(const oVector3& light, oVector3& basisA, oVector3& basisB) throw();
inline oVector3 GroundProjection(const oVector3& v, const oVector3& light) throw();

ulong oVector3Decl[];

class ShadowMesh
{
protected:
	ushort numVertices;
	oHeapPtr<oVector3>::Type vertices;
		
public:
	ShadowMesh();

	void CreateBox(const oVector3& length);
	void CreateCylinder(const oReal& radius, const oReal& length, const ushort& segments);
	void CreateSphere(const oReal& radius, const ushort& segments, const ushort& levels);

	const ushort& GetVertexCount() const throw() { return numVertices; }
	const oVector3& GetVertex(const ushort index) const { return vertices[index]; }
};

class ShadowVolume
{
protected:
	const ShadowMesh* mesh;

	ushort numVertices;
	oHeapPtr<oVector2>::Type hull; // pre-allocated, used for computation not rendering
	oHeapPtr<ushort>::Type hullIndices; // pre-allocated, used for computation not rendering
	oHeapPtr<oVector3>::Type vertices;

public:
	ShadowVolume();

	ShadowVolume(const ShadowMesh* const& mesh_);
	void Create(const ShadowMesh* const& mesh_);
	
	void ComputeShadowVolume(const oMatrix4x4& transform, const oVector3& light, const oVector3& basisA, const oVector3& basisB) throw();
	void ComputeSphereShadowVolumeHack(const oMatrix4x4& transform, const oVector3& light, const oVector3& basisA, const oVector3& basisB, const oReal& radius, const ushort& segments) throw();

	ushort GetHullSize() const throw() { return numVertices; }
	ushort GetVertexCount() const throw() { return 2 * numVertices + 2; }
	ushort GetIndexCount() const throw() { return 4 * numVertices + 6; }

	void GetVertices(oVector3* const& buffer) const throw();
	void GetIndices(ushort* const& buffer, const ushort& offset) const throw();

	void DrawIndexedPrimitive(IDirect3DDevice8* const& device, const ushort& vertexOffset, const ushort& indexOffset) const throw();
};

class GeomShadowVolume : private ShadowVolume
{
protected:
	dGeomID geom;
	ushort vertexOffset, indexOffset;

public:
	GeomShadowVolume();
    
	GeomShadowVolume(const dGeomID& geom_, const ShadowMesh* const& mesh_);
	void Create(const dGeomID& geom_, const ShadowMesh* const& mesh_);

	void ComputeShadowVolume(const oVector3& light, const oVector3& basisA, const oVector3& basisB) throw();
	void ComputeSphereShadowVolumeHack(const oVector3& light, const oVector3& basisA, const oVector3& basisB, const oReal& radius, const ushort& segments) throw();

	template<class Traits> 
	ushort GetVertices(oVertexBuffer<oVector3, Traits>& vertices, const ushort& vertexOffset_)
	{
		vertexOffset = vertexOffset_;
		oVector3* const buffer = vertices.Lock(vertexOffset, GetVertexCount());
		ShadowVolume::GetVertices(buffer);
		vertices.Unlock();
		return vertexOffset + GetVertexCount();
	}


	template<class Traits> 
	ushort GetIndices(oIndexBuffer<ushort, Traits>& indices, const ushort& indexOffset_)
	{
		indexOffset = indexOffset_;
		ushort* const buffer = indices.Lock(indexOffset, GetIndexCount());
		ShadowVolume::GetIndices(buffer, vertexOffset);
		indices.Unlock();
		return indexOffset + GetIndexCount();
	}

	void DrawIndexedPrimitive(IDirect3DDevice8* const& device);
};