#include "openengine.h"

bool oRenderWindow::active = false;
bool oRenderWindow::sizemoved = false;

oRenderWindow::oRenderWindow() : handle(0) {}

oRenderWindow::~oRenderWindow() throw()
{
	if(IsWindow(handle))
	{
		DestroyWindow(handle);
		handle = 0;
	}
}

LRESULT CALLBACK oRenderWindow::RenderWindowProc(HWND window, UINT msg, WPARAM wParam, LPARAM lParam) throw()
{
	switch(msg)
	{
	case WM_KEYDOWN:
		if(wParam == VK_ESCAPE)
			DestroyWindow(window);
		break;

	case WM_EXITSIZEMOVE:
		sizemoved = true;
		break;
	
	case WM_ACTIVATE:
		active = (wParam != WA_INACTIVE);
		break;

	case WM_DESTROY:
		PostQuitMessage(0);
		break;

	case WM_PAINT:
		ValidateRect(window, 0);
		return 0;
	}
	return DefWindowProc(window, msg, wParam, lParam);
}

void oRenderWindow::Create(const HINSTANCE& instance, const _TCHAR* const& title, const long& width, const long& height)
{
	// register the window class
	static const _TCHAR* const renderWindowClass = _T("OpenEngine Render Target Class");
	WNDCLASSEX wcx = {
		sizeof(WNDCLASSEX),
		CS_OWNDC, 
		RenderWindowProc,
		0, 
		0, 
		instance, 
		0, 
		LoadCursor(0, IDC_ARROW), 
		(HBRUSH)(COLOR_WINDOW + 1), 
		0, 
		renderWindowClass,
		0
	};
	if(!RegisterClassEx(&wcx))
		throw _T("oRenderWindow::Create(): RegisterClassEx failed");

	// adjust the window size so that the client area matches the arguments
	RECT size = { 0, 0, width, height };
	AdjustWindowRect(&size, WS_SYSMENU | WS_CAPTION | WS_BORDER, FALSE);
	
	// create the window
	handle = CreateWindowEx(
		0, 
		renderWindowClass, 
		title, 
		WS_SYSMENU | WS_CAPTION | WS_BORDER, 
		CW_USEDEFAULT, 
		0, 
		size.right - size.left, 
		size.bottom - size.top, 
		0, 
		0, 
		instance, 
		0
	);
	if(handle == 0)
		throw _T("oRenderWindow::Create(): CreateWindowEx failed");

	// show the window immediately
	ShowWindow(handle, SW_SHOWNORMAL);
	UpdateWindow(handle);
}

oGraphics::oGraphics() 
  : renderTargetFormat(D3DFMT_UNKNOWN), 
	depthStencilFormat(D3DFMT_UNKNOWN), 
	previousFrameTime(_R(0)), 
	thisFrameTime(_R(0)), 
	performanceUnit(_R(0)) 
{
	memset(&previousFrameCounter, 0, sizeof(LARGE_INTEGER));
}

void oGraphics::GetPerformanceQueryResolution()
{
	LARGE_INTEGER frequency;
	frequency.QuadPart = (LONGLONG)1000;
	if(!QueryPerformanceFrequency(&frequency) || frequency.QuadPart < (LONGLONG)10000)
		throw _T("oGraphics::GetPerformanceQueryResolution(): Insufficient performance query resolution on this system!");
	performanceUnit = _R(1) / _R(frequency.QuadPart);
}

oReal oGraphics::GetFrameTime() throw()
{
	if(previousFrameCounter.QuadPart == (LONGLONG)0)
		QueryPerformanceCounter(&previousFrameCounter);
	else
	{
		QueryPerformanceCounter(&thisFrameCounter);
		thisFrameTime = _R(thisFrameCounter.QuadPart - previousFrameCounter.QuadPart) * performanceUnit;
		memcpy(&previousFrameCounter, &thisFrameCounter, sizeof(LARGE_INTEGER));
		if(thisFramePaused)
			thisFrameTime = previousFrameTime;
		else if(previousFrameTime == _R(0))
			previousFrameTime = thisFrameTime;
		else
			previousFrameTime = thisFrameTime = thisFrameTime * _R(0.05) + previousFrameTime * _R(0.95);
	}
	return thisFrameTime;
}

void oGraphics::SetRenderTargetFormat(const D3DFORMAT& renderTargetFormat_) throw()
{
	renderTargetFormat = renderTargetFormat_;
}

void oGraphics::SetDepthStencilFormat(const D3DFORMAT& depthStencilFormat_) throw()
{
	depthStencilFormat = depthStencilFormat_;
}

void oGraphics::CreateFromLoader(oGraphicsLoader const& loader, const HINSTANCE& instance, const _TCHAR* const& title, const bool& fpuPreserve)
{
	// get direct3d the easy way, from the loader
	direct3d = loader.GetDirect3D();
	direct3d->AddRef();

	// create the target window
	if(loader.IsSelectedWindowed())
		window.Create(instance, title, loader.GetSelectedMode().Width, loader.GetSelectedMode().Height);
	else
		window.Create(instance, title, 200, 100);

	// store the hardware usage flags
	usingHardwareVertexShaders = ((loader.GetSelectedHardwareUsage() & 0x01) != 0);
	usingPixelShaders = ((loader.GetSelectedHardwareUsage() & 0x02) != 0);

	ulong behaviourFlags = (fpuPreserve ? D3DCREATE_FPU_PRESERVE : 0)
		| (usingHardwareVertexShaders ? (D3DCREATE_HARDWARE_VERTEXPROCESSING | D3DCREATE_PUREDEVICE)
									  : D3DCREATE_SOFTWARE_VERTEXPROCESSING);
	vertexUsage = (usingHardwareVertexShaders ? 0 : D3DUSAGE_SOFTWAREPROCESSING); 

	// set up the presentation parameters
	D3DPRESENT_PARAMETERS pp;
	memset(&pp, 0, sizeof(D3DPRESENT_PARAMETERS));
	if(renderTargetFormat == D3DFMT_UNKNOWN) 
		renderTargetFormat = loader.GetSelectedMode().Format;
	pp.BackBufferFormat = renderTargetFormat;
	targetWidth = pp.BackBufferWidth = loader.GetSelectedMode().Width;
	targetHeight = pp.BackBufferHeight = loader.GetSelectedMode().Height;
	pp.BackBufferCount = 2; // TODO: user-defined
	pp.SwapEffect = D3DSWAPEFFECT_DISCARD;
	pp.Windowed = (loader.IsSelectedWindowed() ? TRUE : FALSE);
	if(depthStencilFormat != D3DFMT_UNKNOWN)
	{
		pp.EnableAutoDepthStencil = TRUE;
		pp.AutoDepthStencilFormat = depthStencilFormat;
	}

	// compute the aspect ratio
	targetAspectRatio = _R(4) / _R(3);
	if(loader.IsSelectedWindowed())
	{
		D3DDISPLAYMODE currentMode;
		HRESULT hr = direct3d->GetAdapterDisplayMode(loader.GetSelectedAdapter(), &currentMode);
		if(FAILED(hr))
			throw _T("oGraphics::CreateFromLoader(): GetAdapterDisplayMode failed");
		targetAspectRatio	*= _R(targetWidth * currentMode.Height) / _R(targetHeight * currentMode.Width);
	}

	// create the device
	HRESULT hr = direct3d->CreateDevice(loader.GetSelectedAdapter(), D3DDEVTYPE_HAL, window, behaviourFlags, &pp, &device);
	if(FAILED(hr))
		throw _T("oGraphics::CreateFromLoader(): CreateDevice failed");
}

bool oGraphics::thisFramePaused = false;

bool oGraphics::PurgeMessageQueue() throw()
{
	MSG msg;
	thisFramePaused = false;
	for(;;)
	{
		// purge ye olde message queue
		while(PeekMessage(&msg, 0, 0, 0, PM_REMOVE))
		{
			// quit on WM_QUIT
			if(msg.message == WM_QUIT)
				return true;
			// otherwise process normally
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
		// only return if active
		if(oRenderWindow::IsActive())
		{
			thisFramePaused = thisFramePaused || oRenderWindow::WasSizeMoved();
			return false;
		}
		// otherwise wait until we're active
		thisFramePaused = true;
		WaitMessage();
	}
}

void oGraphics::ReleaseStreamSource(const ulong& num)
{
	device->SetStreamSource(num, 0, 0);
}

void oGraphics::CreateVertexShader(oVertexShader& shader, ulong* const& decl, ulong* const& body)
{
	// create the shader
	ulong handle;
	HRESULT hr = device->CreateVertexShader(decl, body, &handle, vertexUsage);
	if(FAILED(hr))
		throw _T("oGraphics::CreateVertexShader(): CreateVertexShader failed");
	// store the result
	shader.Create(device, handle);
}

void oGraphics::CreatePixelShader(oPixelShader& shader, ulong* const& function)
{
	// create the shader
	ulong handle;
	HRESULT hr = device->CreatePixelShader(function, &handle);
	if(FAILED(hr))
		throw _T("oGraphics::CreatePixelShader(): CreatePixelShader failed");
	// store the result
	shader.Create(device, handle);
}

IDirect3DTexture8* oGraphics::LoadTexture(const _TCHAR* const& fileName, const ulong& mipLevels, const ulong& usage, const D3DFORMAT& format, const D3DPOOL& pool)
{
	IDirect3DTexture8* tmp = 0;
	HRESULT hr = D3DXCreateTextureFromFileEx(
		device, 
		fileName, 
		D3DX_DEFAULT, 
		D3DX_DEFAULT, 
		mipLevels, 
		usage, 
		format, 
		pool, 
		D3DX_DEFAULT, D3DX_DEFAULT, 0, 0, 0, 
		&tmp
	);
	if(FAILED(hr))
		throw _T("oGraphics::LoadTexture(): D3DXCreateTextureFromFileEx failed!");
	return tmp;
}
