#include "openengine.h"

INT_PTR CALLBACK oGraphicsLoader::DialogProc(HWND dialog, UINT msg, WPARAM wParam, LPARAM lParam) throw()
{
	try
	{
		switch(msg)
		{
		case WM_INITDIALOG:
			{
				// get and store the calling instance
				if(!lParam)
					throw _T("oGraphicsLoader::DialogProc(): Invalid this pointer passed during WM_INITDIALOG");
				oGraphicsLoader& thisLoader = *reinterpret_cast<oGraphicsLoader*>(lParam);
#pragma warning(disable: 4244)
				SetWindowLongPtr(dialog, GWLP_USERDATA, reinterpret_cast<LONG_PTR>(&thisLoader));
#pragma warning(default: 4244)
				// load up the adapters on this system
				thisLoader.LoadAdapters(dialog);
				thisLoader.adapter = D3DADAPTER_DEFAULT;
				const _TCHAR* failure = thisLoader.CheckAdapter();
				// load up the hardware usage options
				thisLoader.LoadHardwareOptions(dialog);
				// default to windowed
				thisLoader.windowed = true;
				SendDlgItemMessage(dialog, IDC_WINDOWED, BM_SETCHECK, BST_CHECKED, 0);
				// update the modes
				thisLoader.UpdateModes(dialog);
				// show any errors if they occurred
				if(failure)
				{
					ShowWindow(dialog, SW_SHOW);
					UpdateWindow(dialog);
					MessageBox(dialog, failure, _T("OpenEngine: Adapter failed"), MB_OK | MB_ICONERROR | MB_TASKMODAL);
					EnableWindow(GetDlgItem(dialog, IDOK), !failure);
					EnableWindow(GetDlgItem(dialog, IDC_HARDWAREUSAGE), !failure);
					EnableWindow(GetDlgItem(dialog, IDC_WINDOWED), !failure);
					EnableWindow(GetDlgItem(dialog, IDC_MODES), !failure);
				}
				return TRUE;
			}

		case WM_COMMAND:
			switch(wParam)
			{
			case ((CBN_SELCHANGE << 16) | IDC_ADAPTER):
				{
					// get the loader
					oGraphicsLoader& thisLoader = *GetGraphicsLoader(dialog);

					// update the adapter and modes
					thisLoader.adapter = static_cast<ulong>(SendDlgItemMessage(dialog, IDC_ADAPTER, CB_GETCURSEL, 0, 0));
					const _TCHAR* failure = thisLoader.CheckAdapter();
					if(failure)
						MessageBox(dialog, failure, _T("OpenEngine: Adapter failed"), MB_OK | MB_ICONERROR | MB_TASKMODAL);

					EnableWindow(GetDlgItem(dialog, IDOK), !failure);
					EnableWindow(GetDlgItem(dialog, IDC_HARDWAREUSAGE), !failure);
					EnableWindow(GetDlgItem(dialog, IDC_WINDOWED), !failure);
					EnableWindow(GetDlgItem(dialog, IDC_MODES), !failure);

					// load up the hardware usage options and display modes
					thisLoader.LoadHardwareOptions(dialog);
					thisLoader.UpdateModes(dialog);
					return TRUE;
				}

			case ((BN_CLICKED << 16) | IDC_WINDOWED):
				{
					// get the loader
					oGraphicsLoader& thisLoader = *GetGraphicsLoader(dialog);

					// update the windowedness
					thisLoader.windowed = (SendDlgItemMessage(dialog, IDC_WINDOWED, BM_GETCHECK, 0, 0) == BST_CHECKED);
					thisLoader.UpdateModes(dialog);
					return TRUE;
				}				

			case IDOK:
				{
					// get the loader
					oGraphicsLoader& thisLoader = *GetGraphicsLoader(dialog);

					// save the hardware usage
					thisLoader.selectedHardwareUsage = static_cast<uchar>(SendDlgItemMessage(dialog, IDC_HARDWAREUSAGE, CB_GETITEMDATA, 
						SendDlgItemMessage(dialog, IDC_HARDWAREUSAGE, CB_GETCURSEL, 0, 0), 0));

					// save the mode
					HRESULT hr = thisLoader.direct3d->EnumAdapterModes(
						thisLoader.adapter, 
						static_cast<ulong>(SendDlgItemMessage(dialog, IDC_MODES, LB_GETITEMDATA, 
							SendDlgItemMessage(dialog, IDC_MODES, LB_GETCURSEL, 0, 0), 0)), 
						&thisLoader.selectedMode
					);
                    if(FAILED(hr))
						throw _T("oGraphicsLoader::DialogProc(): EnumAdapterModes failed");

					// close the dialog
					EndDialog(dialog, 0);
					return TRUE;
				}

			case IDCANCEL:
				// cancel the dialog
				EndDialog(dialog, 1);
				return TRUE;
			}
			break;

		case WM_CLOSE:
			// cancel the dialog
			EndDialog(dialog, 2);
			return TRUE;
		}
	}
	catch(_TCHAR* excuse)
	{
		MessageBox(dialog, excuse, _T("OpenEngine: Commencing harakiri"), MB_OK | MB_ICONERROR | MB_TASKMODAL);
		// cancel the dialog
		EndDialog(dialog, 3);
		return TRUE;
	}
	// default to the standard message processor
	return FALSE;
}

oGraphicsLoader::oGraphicsLoader() 
  : minVertexShaderVersion(0x0100), 
    minPixelShaderVersion(0x0101), 
	supportLegacyPixelPipeline(true), 
	adapter(0), 
	windowed(true), 
	hardwareVertexShaders(false), 
	hardwarePixelShaders(false),
	selectedHardwareUsage(0)
{
}

void oGraphicsLoader::SetMinVertexShaderVersion(const ushort& version) throw()
{
	if(version >= 0x0100)
		minVertexShaderVersion = version;
}

void oGraphicsLoader::SetMinPixelShaderVersion(const ushort& version) throw()
{
	if(version >= 0x0101)
		minPixelShaderVersion = version;
}

void oGraphicsLoader::SupportLegacyPixelPipeline(const bool& support) throw()
{
	supportLegacyPixelPipeline = support;
}

const _TCHAR* oGraphicsLoader::CheckAdapter()
{
	// get the device caps
	D3DCAPS8 caps;
	HRESULT hr = direct3d->GetDeviceCaps(adapter, D3DDEVTYPE_HAL, &caps);
	if(FAILED(hr))
		throw _T("oGraphicsLoader::CheckAdapter(): GetDeviceCaps failed");
	
	// check the vertex and pixel shader versions
    caps.VertexShaderVersion &= 0xffff; // the top 2 bytes usually contain garbage
	hardwareVertexShaders = (caps.VertexShaderVersion >= minVertexShaderVersion);
    caps.PixelShaderVersion &= 0xffff; // the top 2 bytes usually contain garbage
	hardwarePixelShaders = (caps.PixelShaderVersion >= minPixelShaderVersion);
	if(!supportLegacyPixelPipeline)
	{
		if(!hardwareVertexShaders)
			return _T("This device does not meet the minimum vertex shader version requirements!");
		if(!hardwarePixelShaders)
			return _T("This device does not meet the minimum pixel shader version requirements!");
	}

	// adapter ok
	return 0;
}

void oGraphicsLoader::LoadHardwareOptions(const HWND& dialog) throw()
{
	// empty the current list
	SendDlgItemMessage(dialog, IDC_HARDWAREUSAGE, CB_RESETCONTENT, 0, 0);

	// add in the options
	if(hardwarePixelShaders && hardwareVertexShaders)
	{
		LRESULT index = SendDlgItemMessage(dialog, IDC_HARDWAREUSAGE, CB_ADDSTRING, 0, reinterpret_cast<LPARAM>(_T("Vertex & Pixel Shaders")));
		SendDlgItemMessage(dialog, IDC_HARDWAREUSAGE, CB_SETITEMDATA, index, 0x03);
	}
	if(hardwareVertexShaders && supportLegacyPixelPipeline)
	{
		LRESULT index = SendDlgItemMessage(dialog, IDC_HARDWAREUSAGE, CB_ADDSTRING, 0, reinterpret_cast<LPARAM>(_T("Vertex Shaders")));
		SendDlgItemMessage(dialog, IDC_HARDWAREUSAGE, CB_SETITEMDATA, index, 0x01);
	}
	if(supportLegacyPixelPipeline)
	{
		LRESULT index = SendDlgItemMessage(dialog, IDC_HARDWAREUSAGE, CB_ADDSTRING, 0, reinterpret_cast<LPARAM>(_T("Software Shaders Only")));
		SendDlgItemMessage(dialog, IDC_HARDWAREUSAGE, CB_SETITEMDATA, index, 0x00);
	}

	// select the current one
	SendDlgItemMessage(dialog, IDC_HARDWAREUSAGE, CB_SETCURSEL, 0, 0);
}

oGraphicsLoader* oGraphicsLoader::GetGraphicsLoader(const HWND& dialog)
{
	// get the loader from the long pointer associated with the window
	LONG_PTR temp = GetWindowLongPtr(dialog, GWLP_USERDATA);
	if(!temp)
		throw _T("oGraphicsLoader::DialogProc(): GetWindowLongPtr returned 0");
	return reinterpret_cast<oGraphicsLoader*>(temp);
}

void oGraphicsLoader::LoadAdapters(const HWND& dialog)
{
	// empty the current list
	SendDlgItemMessage(dialog, IDC_ADAPTER, CB_RESETCONTENT, 0, 0);

	// get the number of adapters
	ulong numAdapters = direct3d->GetAdapterCount();
	if(!numAdapters)
		throw _T("oGraphicsLoader::LoadAdapters(): GetAdapterCount returned 0");
	
	// load in each adapter
	_TCHAR buffer[2 * MAX_DEVICE_IDENTIFIER_STRING];
	for(ulong i = 0; i < numAdapters; ++i)
	{
		D3DADAPTER_IDENTIFIER8 ident;
		HRESULT hr = direct3d->GetAdapterIdentifier(i, D3DENUM_NO_WHQL_LEVEL, &ident);
		if(FAILED(hr))
            throw _T("oGraphicsLoader::LoadAdapters(): GetAdapterIdentifier failed");

		_stprintf(buffer, _T("%hs (%hs)"), ident.Driver ,ident.Description);
		SendDlgItemMessage(dialog, IDC_ADAPTER, CB_ADDSTRING, 0, reinterpret_cast<LPARAM>(buffer));
	}

	// select the current one
	SendDlgItemMessage(dialog, IDC_ADAPTER, CB_SETCURSEL, adapter, 0);
}

bool oGraphicsLoader::CompareMode(IndexedDisplayMode* mode1, IndexedDisplayMode* mode2) throw()
{
	// sort by format, then mode area, then refresh rate
	if(mode1->mode.Format != mode2->mode.Format)
		return (mode1->mode.Format < mode2->mode.Format);

	if((mode1->mode.Width * mode1->mode.Height) != (mode2->mode.Width * mode2->mode.Height))
		return (mode1->mode.Width * mode1->mode.Height) > (mode2->mode.Width * mode2->mode.Height);

    return (mode1->mode.RefreshRate > mode2->mode.RefreshRate);
}

void oGraphicsLoader::UpdateModes(HWND const& dialog)
{
	// add all the modes this adapter supports
	ulong numModes = direct3d->GetAdapterModeCount(adapter);
	if(!numModes)
		throw _T("oGraphicsLoader::UpdateModes(): GetAdapterModeCount returned 0");

    oArrayPtr<IndexedDisplayMode>::Type modes = new IndexedDisplayMode[numModes];
	std::vector<IndexedDisplayMode*> sorted(numModes);
	for(ulong i = 0; i < numModes; ++i)
	{
		HRESULT hr = direct3d->EnumAdapterModes(adapter, i, &modes[i].mode);
		if(FAILED(hr))
			throw _T("oGraphicsLoader::UpdateModes(): EnumAdapterModes failed");
		modes[i].index = i;
		sorted[i] = &modes[i];
	}

	// sort the modes 
	std::sort(sorted.begin(), sorted.end(), CompareMode);

	// get the current display mode for this adapter
	D3DDISPLAYMODE current;
	HRESULT hr = direct3d->GetAdapterDisplayMode(adapter, &current);
	if(FAILED(hr))
		throw _T("oGraphicsLoader::UpdateModes(): GetAdapterDisplayMode failed");

	// empty the current list
	SendDlgItemMessage(dialog, IDC_MODES, LB_RESETCONTENT, 0, 0);
	
	// chuck them into the list
	// corresponding formats: D3DFMT_A8R8G8B8 = 21, D3DFMT_X8R8G8B8, D3DFMT_R5G6B5, D3DFMT_X1R5G5B5
	static _TCHAR* bpp[] = { _T("32"), _T("8+24"), _T("16"), _T("15") };
	_TCHAR buffer[50];
	for(std::vector<IndexedDisplayMode*>::const_iterator it = sorted.begin(); it != sorted.end(); ++it)
	{
        if(!windowed || ((*it)->mode.RefreshRate == current.RefreshRate
			&& ((*it)->mode.Format == current.Format)
			&& ((*it)->mode.Width < current.Width)
			&& ((*it)->mode.Height < current.Height)))
		{
			_stprintf(buffer, _T("%dx%d %dHz (%sbpp)"), 
				(*it)->mode.Width, (*it)->mode.Height, (*it)->mode.RefreshRate, bpp[(*it)->mode.Format - D3DFMT_A8R8G8B8]);
			LRESULT index = SendDlgItemMessage(dialog, IDC_MODES, LB_ADDSTRING, 0, (LPARAM)buffer);
			SendDlgItemMessage(dialog, IDC_MODES, LB_SETITEMDATA, index, (*it)->index);
		}
	}

	// select the first mode in the list
	SendDlgItemMessage(dialog, IDC_MODES, LB_SETCURSEL, 0, 0);
}

bool oGraphicsLoader::Show(HINSTANCE instance)
{
	// make sure we're got a Direct3D8 object
	if(!direct3d && !(direct3d = Direct3DCreate8(D3D_SDK_VERSION)))
		throw _T("oGraphicsLoader::Show(): Direct3DCreate8 failed");

	// show the dialog
#pragma warning(disable: 4311)
	INT_PTR result = DialogBoxParam(instance, MAKEINTRESOURCE(IDD_LOADER), 0, DialogProc, reinterpret_cast<DWORD>(this));
#pragma warning(default: 4311)

	// return true if the user clicked on OK with a valid mode
	return (result == 0);
}
