#pragma once

#define oRealSqrt(x) sqrtf(x)
#define oRealAbs(x) fabsf(x)
#define oRealCos(x) cosf(x)
#define oRealSin(x) sinf(x)

#define oRealInvSin(x) asinf(x)

#define oRealPi 3.14159265358979323f
#define oRealEpsilon FLT_EPSILON

class oVector3
{
public:
	oReal x, y, z;

public:
	oVector3() throw();

	oVector3(const oReal& x_, const oReal& y_, const oReal& z_) throw();

	oVector3(const oVector3& v) throw();
	oVector3& operator=(const oVector3& v) throw();

#ifdef OPENENGINE_USING_ODE	
	explicit oVector3(const dReal* const& v) { CreateFromdVector3(v); }
	inline void CreateFromdVector3(const dReal* const& v) throw() { x = _R(v[0]); y = _R(v[1]); z = _R(v[2]); }
#endif // def OPENENGINE_USING_ODE	

	oReal Norm() const throw();
	oReal NormSquared() const throw();

	oVector3 operator-() const throw();

	const oVector3& operator+=(const oVector3& v) throw();
	const oVector3& operator-=(const oVector3& v) throw();
	const oVector3& operator*=(const oVector3& v) throw();

	const oVector3& operator*=(const oReal& r) throw();
	const oVector3& operator/=(const oReal& r) throw();
};

oVector3 operator+(const oVector3& l, const oVector3& r) throw();
oVector3 operator-(const oVector3& l, const oVector3& r) throw();

oVector3 operator*(const oVector3& l, const oVector3& r) throw();
oVector3 operator*(const oReal& l, const oVector3& r) throw();
oVector3 operator*(const oVector3& l, const oReal& r) throw();
oVector3 operator/(const oVector3& l, const oReal& r) throw();

oReal oDotProduct(const oVector3& l, const oVector3& r) throw();
oVector3 oCrossProduct(const oVector3& l, const oVector3& r) throw();

class oLine3
{
public:
	oVector3 p, d;

public:
	oLine3() throw();

	oLine3(const oVector3& p_, const oVector3& d_) throw();
};

class oMatrix4x4;

class oQuaternion
{
public:
	oReal w, x, y, z;

	oQuaternion() throw();
	oQuaternion(const oReal& w_, const oReal& x_, const oReal& y_, const oReal& z_) throw();

#ifdef OPENENGINE_USING_ODE	
	explicit oQuaternion(const dReal* const& q) throw() { CreateFromdQuaternion(q); }
	void CreateFromdQuaternion(const dReal* const& q) throw() { w = _R(q[0]); x = _R(q[1]); y = _R(q[2]); z = _R(q[3]); }
#endif // def OPENENGINE_USING_ODE	

	oQuaternion(const oVector3& axis, const oReal& angle) throw();
	void CreateFromAxisAngle(const oVector3& axis, const oReal& angle) throw();

	explicit oQuaternion(const oVector3& lookat) throw();
	void CreateFromLookAtPoint(const oVector3& lookat) throw();

	explicit oQuaternion(const oMatrix4x4& m) throw();
	void CreateFromMatrix(const oMatrix4x4& m) throw();

	oReal Norm() const throw();
	oReal NormSquared() const throw();
	
	oQuaternion operator~() const throw();

	const oQuaternion& operator/=(const oReal& r);
};

oQuaternion operator*(const oQuaternion& l, const oQuaternion& r) throw();
oQuaternion operator/(const oQuaternion& l, const oReal& r) throw();

oVector3 oQuaternionRotation(const oQuaternion& q, const oVector3& v) throw();
oQuaternion oQuaternionPower(const oQuaternion& q, const oReal& t) throw(); // assumes a unit quaternion
oQuaternion oQuaternionSLERP(const oQuaternion& a, const oQuaternion& b, const oReal& t) throw();

class oMatrix4x4
{
public:
	oReal e11, e12, e13, e14;
	oReal e21, e22, e23, e24;
	oReal e31, e32, e33, e34;
	oReal e41, e42, e43, e44;

	oMatrix4x4() throw();
	oMatrix4x4(const oReal& e11_, const oReal& e12_, const oReal& e13_, const oReal& e14_, 
			   const oReal& e21_, const oReal& e22_, const oReal& e23_, const oReal& e24_, 
			   const oReal& e31_, const oReal& e32_, const oReal& e33_, const oReal& e34_, 
			   const oReal& e41_, const oReal& e42_, const oReal& e43_, const oReal& e44_) throw();
	oMatrix4x4(const oReal& e11_, const oReal& e12_, const oReal& e13_, 
			   const oReal& e21_, const oReal& e22_, const oReal& e23_, 
			   const oReal& e31_, const oReal& e32_, const oReal& e33_) throw();
	void CreateIdentity() throw();

	oMatrix4x4(const oMatrix4x4& m) throw();
	oMatrix4x4& operator=(const oMatrix4x4& m) throw();

	explicit oMatrix4x4(const oQuaternion& q) throw();
	void CreateFromQuaternion(const oQuaternion& q) throw();

#ifdef OPENENGINE_USING_ODE	
	explicit oMatrix4x4(const dReal* const& m) throw() { CreateFromdMatrix3(m); }
	void CreateFromdMatrix3(const dReal* const& m) throw()
	{
		e11 = _R(m[0]); e12 = _R(m[1]); e13 = _R(m[2]);
		e21 = _R(m[4]); e22 = _R(m[5]); e23 = _R(m[6]);
		e31 = _R(m[8]); e32 = _R(m[9]); e33 = _R(m[10]);
		e14 = e24 = e34 = e41 = e42 = e43 = _R(0);
		e44 = _R(1);
	}
#endif // def OPENENGINE_USING_ODE	

	void CreateFromScale(const oVector3& v) throw();

	void PreTranslate(const oVector3& v) throw();
	void PostTranslate(const oVector3& v) throw();

	void PreScale(const oVector3& v) throw();
	void PostScale(const oVector3& v) throw(); 

	oMatrix4x4 Transpose() const throw();

	oVector3 GetTranslation() const throw();
};

oMatrix4x4 operator*(const oMatrix4x4& l, const oMatrix4x4& r) throw();

oVector3 operator*(const oMatrix4x4& l, const oVector3& r) throw();
oVector3 oTransformAs3x3(const oMatrix4x4& l, const oVector3& r) throw();

class oProjection
{
protected:
	oMatrix4x4 transform;

public:
	oReal fovRatio; //!< Distance/half-width
	oReal aspectRatio; 
	oReal zNear;
	oReal zFar;

	oProjection() throw();

	void CreateTransform() throw();
	const oMatrix4x4& GetTransform() const throw() { return transform; }
};

class oVector2
{
public:
	oReal x, y;

	oVector2() throw();
	oVector2(const oReal& x_, const oReal& y_) throw();

	oVector2(const oVector2& v) throw();
	oVector2& operator=(const oVector2& v) throw();

	oReal Norm() const throw();
	oReal NormSquared() const throw();

	const oVector2& operator*=(const oReal& r) throw();
	const oVector2& operator/=(const oReal& r) throw();

	const oVector2& operator*=(const oVector2& v) throw();
};

oVector2 operator-(const oVector2& l, const oVector2& r) throw();

oReal oDotProduct(const oVector2& l, const oVector2& r) throw();

class oPlane2
{
public:
	oVector2 n;
	oReal d;

	oPlane2() throw();
	oPlane2(const oVector2& n_, const oReal& d) throw();

	oPlane2(const oPlane2& l) throw();
	oPlane2& operator=(const oPlane2& l) throw();

	oPlane2(const oVector2& a, const oVector2& b) throw();
	void CreateFromPoints(const oVector2& a, const oVector2& b) throw();

	oReal PointHeight(const oVector2& v) const throw();
	bool PointAboveTol2(const oVector2& v, const oReal& tol2) const throw();
};

ushort oQuickHull(const oVector2* const& points, const ulong& count, const oReal& toleranceSq, ushort* const& hull) throw();

void oCreatePlanTransform(const oVector2& min, const oVector2& max, const oReal& y, oMatrix4x4& transform) throw();
void oCreatePlanPreserveAspect(oVector2& min, oVector2& max, const oReal& aspectRatio) throw();

class oPlane3
{
public:
	oVector3 n;
	oReal d;

	oPlane3() throw();
	oPlane3(const oVector3& n_, const oReal& d_) throw();

	oPlane3(const oVector3& a, const oVector3& b, const oVector3& c) throw();
	void CreateFromPoints(const oVector3& a, const oVector3& b, const oVector3& c) throw();

	oReal PointHeight(const oVector3& v) const throw();

	bool IntersectsAABB(const oVector3& centre, const oVector3& halfLengths) const throw();
};

class oFrustrum
{
public:
	oVector3 points[5];
	oPlane3 bounds[6];

	oFrustrum(const oVector3& position, const oQuaternion& orientation, const oProjection& projection) throw();

	bool IntersectsAABB(const oVector3& centre, const oVector3& halfLengths) const throw();
};